#pragma once

#include <cassert>
#include <chrono>
#include <ostream>
#include <sstream>
#include <vector>
#include <exception>
#include <cmath>
#include <iomanip>



template<typename TYPE>
std::ostream& operator<<(std::ostream& out, const std::chrono::duration<TYPE>& input){
    TYPE T = std::chrono::duration_cast<std::chrono::nanoseconds>(input).count();
    out.precision(0);
    out << std::fixed;
    out << std::setfill('0');
        out << std::setw(3);
    if (T < 1000.) {
        out << T << "ns";
    } else if ( (T/=1000.) < 1000. ) {
        out << T << "μs";
    } else if ( (T/=1000.) < 1000. ) {
        out << T << "ms";
    } else if ( (T/=1000.) < 60. ) {
        if (T<10) out.precision(1);
        out << std::setw(1);
        out << T << "s";
    } else if ( (T/=60.) < 60. ) {
        out << std::setw(0);
        out << std::round(T) << "min";
        out << std::setw(2);
        out << std::round((size_t)(T*60.) % 60) << "s";
    } else {
        out << std::setw(0);
        out.precision(0);
        out << std::floor(T/60.) << "h" << ((size_t)T*100 % 6000)/100 << "min";
    }

    return out;
}

template<typename TYPE = double>
struct measure_time {
    size_t count = 0;
    measure_time(std::string _description = "") : description(_description) {}

    void start(){
#pragma omp master
        {
            assert(active==false && "A time measurement is already running.");
            begin = std::chrono::high_resolution_clock::now();
            active = true;
        }
    }

    auto stop(){
#pragma omp master
        {
            assert(active && "The measurement has not been started.");
            last = std::chrono::high_resolution_clock::now() - begin;
            duration+=last;
            count++;
            active=false;
        }
        return last;
    }

    auto get_last(){
        return last;
    }
    auto get_mean() {
        return duration/count;
    }
    auto get_duration() {
        return duration;
    }

    std::string print() const {
        std::ostringstream output;

        output << "Time_measurement";
        output << " " << description;
        std::chrono::duration<TYPE> tmp = (count>0)?(duration/count):duration;
        output << " " << " t_total=" << duration << "(" << count << "*" << tmp << ")";

        return output.str();
    }

protected:
    std::string description;
    bool active = false;
    std::chrono::high_resolution_clock::time_point begin;
    std::chrono::duration<TYPE> duration = std::chrono::nanoseconds::zero();
    std::chrono::duration<TYPE> last = std::chrono::nanoseconds::zero();
};

template<typename TYPE>
std::ostream& operator<<(std::ostream& out, const measure_time<TYPE>& input){
    out << input.print();

    return out;
}


std::string io_from_rhs_time (const std::chrono::duration<realtype>& t_rhs) {
    extern const size_t Mx, MVn, MVp;

    realtype val=2.*MX*MVn*MVp*sizeof(realtype)/std::chrono::duration_cast<std::chrono::nanoseconds>(t_rhs).count() * 1e9;

    std::ostringstream output;
    output.precision(2);
    output << std::fixed;
    if (val<std::pow(2,10)) {
        output << val << "B/s";
    } else if (val<std::pow(2,20)) {
        val /= std::pow(2,10);
        output << val << "KB/s";
    } else if (val<std::pow(2,30)) {
        val /= std::pow(2,20);
        output << val << "MB/s";
    } else if (val<std::pow(2,40)) {
        val /= std::pow(2,30);
        output << val << "GB/s";
    } else {
        val /= std::pow(2,40);
        output << val << "TB/s";
    }
    return output.str();
}
