/* -----------------------------------------------------------------
 * Based upon the cvsAdvDiff_bnd.c example.
 * -----------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include <fstream>
#include <string>
#include <sstream>
#include <ostream>
#include <iostream>
#include "constants.h"

/* Header files with a description of contents used */

#include <cvodes/cvodes.h>             /* prototypes for CVODES fcts., consts. */
#include <sunmatrix/sunmatrix_band.h>  /* access to band SUNMatrix             */
#include <sunlinsol/sunlinsol_band.h>  /* access to band SUNLinearSolver       */
#include <cvodes/cvodes_direct.h>      /* access to CVDls interface            */
#include <sundials/sundials_types.h>   /* definition of type realtype          */
#include <sundials/sundials_math.h>    /* definition of SUNRabs and SUNRexp    */
#include <sunlinsol/sunlinsol_spgmr.h>
#include <cvode/cvode_spils.h>

#include "boundary.h"

// using size_t = long;

/* Problem Constants */
const std::string output_path = "./output/nominalVT_pseudocyl_movie/";

realtype deviation(const realtype& input){
    return 1.;
}

#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#define MIN(x, y) (((x) < (y)) ? (x) : (y))

constexpr size_t order = 3;

constexpr realtype m_ion = m_p;

const realtype vn_T_e = v_thermal_eV(1.5e3,m_e);
const realtype vp_T_e = v_thermal_eV(1.5e3,m_e);
const realtype vn_T_i = v_thermal_eV(1.5e3,m_ion);
const realtype vp_T_i = v_thermal_eV(1.5e3,m_ion);

const realtype v_normalization = vp_T_i;

constexpr realtype n_0 = 5e19;
constexpr realtype n_normalization = n_0;
// constexpr realtype E_min_x = 0; // currently an arbitrary number

constexpr btype boundary_x = btype::DIRICHLET; // otherwise double Maxwellian DBC
constexpr btype boundary_vn = btype::PERIODIC;
constexpr btype boundary_vp = btype::PERIODIC;

constexpr size_t MX = 128;             /* mesh dimensions           */
constexpr size_t MVn = 48;
constexpr size_t MVp = 128;

#include "measure_time.h" //NOTE: io_from_rhs_time in meawsure_time.h depends on the MX, MVp and MVn.

constexpr realtype XMIN  = -5; // do not write here
constexpr realtype XMAX  = 5;    /* domain boundaries         */
constexpr realtype X_LENGTH = XMAX-XMIN;
constexpr size_t dim_v_n = 2;
constexpr realtype domain_thermal_factor = 5.;
constexpr realtype DX = X_LENGTH/((realtype) MX+(boundary_x!=btype::PERIODIC) );
constexpr realtype XSTART = XMIN+((boundary_x!=btype::PERIODIC)*DX);
static realtype x_by(size_t i) { return XSTART + i*DX;}

constexpr realtype artificial_e_mass_multiplier = 1.;
constexpr realtype const_colrate = .0;
const std::string filename = "f";

// // unit-free normalization
constexpr realtype unit_m = m_p;

constexpr size_t NEQ = MX*MVn*MVp;          /* number of equations       */
constexpr realtype ATOL = 1.0e-10; /* scalar absolute tolerance */
constexpr realtype RTOL = 1e-7; /* scalar absolute tolerance */
constexpr realtype T0 = 0.0;    /* initial time              */
constexpr realtype T1 = 10.0;    /* first output time         */
constexpr realtype DTOUT = 5e-3;    /* output time increment     */
constexpr size_t total_steps = (T1-T0)/DTOUT;
constexpr size_t NOUT = 1e5;             /* max number of output times    */

constexpr realtype lambda = 1.;
const realtype lambda_D_0 = std::sqrt(epsilon_0*unit_m*std::pow(v_normalization/q_e,2)/n_normalization/2.);
const realtype tau = lambda_D_0/v_normalization;

inline constexpr static realtype& IJKth(realtype* vdata, size_t x, size_t vn, size_t vp) {
    assert(x>=0);
    assert(x<MX);
    assert(vp>=0);
    assert(vp<MVp);
    assert(vn>=0);
    assert(vn<MVn);
    return vdata[(vp) + (vn)*MVp + (x)*MVp*MVn];
}
inline constexpr static realtype* pIJKth(realtype* vdata, size_t x, size_t vn, size_t vp) {
    assert(x>=0);
    assert(x<MX);
    assert(vp>=0);
    assert(vp<MVp);
    assert(vn>=0);
    assert(vn<MVn);
    return &vdata[(vp) + (vn)*MVp + (x)*MVp*MVn];
}

measure_time<realtype> mt_total("total");
measure_time<realtype> mt_cvode_i("cvode (ions)");
measure_time<realtype> mt_cvode_e("cvode (electrons)");
measure_time<realtype> mt_integrator("integrator");
measure_time<realtype> mt_efield("Efield");
measure_time<realtype> mt_f("f");

/* Type : UserData (contains grid constants) */
#include "userdata.h"

/* Private Helper Functions */
#include "sundial_stuff.h"

/* Functions Called by the Solver */
#include "vlasov_collisionless.h"
#include "initial_conditions.h"
#include "electric_field.h"
#include "export_netcdf.h"
#include "collision_operators.h"

#include "openmp_nv_overlay.h"

#include "cvode/cvode_impl.h"

bool one_step(const realtype& tout, N_Vector& q, N_Vector& E, std::vector<realtype>& electric_energy, UserData& e, UserData& i){
    extern measure_time<realtype> mt_integrator;

    int flag = 0;
    bool stop = false;

    //         std::cerr << "solve the i distribution"  << std::endl;
    mt_cvode_i.start();
    flag = CVode ( i.cvode_mem, tout, i.f, &i.time, CV_NORMAL );
    if ( check_flag ( &flag, "CVode ions", 1 ) ) {
        std::cerr << "CVode failed with the ion computation. step=" << i.cvodememrec()->cv_h << std::endl;

        std::cerr << "vp_i\tvn_i\tn" << std::endl;
        for(size_t iii=0; iii<MX; ++iii){
            std::cerr.precision(8);
            std::cerr << i.v_thermal_p[iii] << "\t" << i.v_thermal_n[iii] << "\t" << i.n[iii] << std::endl;
        }

        return false;
    }
    mt_cvode_i.stop();

//     (CVodeMemRec*) i.cvode_mem->cv_hmin;
    //         std::cerr << "solve the e distribution" << std::endl;
    mt_cvode_e.start();
    flag = CVode ( e.cvode_mem, tout, e.f, &e.time, CV_NORMAL );
    if ( check_flag ( &flag, "CVode electrons", 1 ) ) {
        std::cerr << "CVode failed with the electron computation. step=" << e.cvodememrec()->cv_h << std::endl;

        std::cerr << "vp_e\tvn_e" << std::endl;
        for(size_t iii=0; iii<MX; ++iii){
            std::cerr.precision(8);
            std::cerr << e.v_thermal_p[iii] << "\t" << e.v_thermal_n[iii] << "\t" << i.n[iii] << std::endl;
        }

        return false;
    }
    mt_cvode_e.stop();

    // //         update n, Tn, Tp for e and i. And then update the total charge density and consequently the electric field..
    mt_integrator.start();
    i.update_n();
    e.update_n();
    mt_integrator.stop();
    q_from_ns(q, e, i);
    mt_efield.start();
    electric_energy.push_back(E_from_q(E,q));
    mt_efield.stop();

    return true;
}




/*
 * -------------------------------
 * Main Program
 *-------------------------------
 */


int main ( void )
{
    std::cerr << "project2 build on " << __DATE__ << " at " << __TIME__ << std::endl;

// The recurrence time is given by t_r = 2pi/(k *Dvp). with k=2pi/L, where L is the spatial domain size, which normalized to 1. Hence, one gets t_r=1/DVp=MVp
//     std::cerr << "Recurrence time: " << MVp << std::endl;

    omp_message();

// //     define general variables
    N_Vector q, E;
    std::vector<realtype> electric_energy;
    SUNLinearSolver LS_i, LS_e;
    int flag;
    int cv_order_i;


    //NOTE: here first touch could be beneficial.
// //   f_ion
    LS_i = NULL;

    UserData ions(m_ion, 1., n_0, vp_T_i, vn_T_i, const_colrate);
    ions.description="ion distribution";

// // f_e
    LS_e = NULL;

    UserData electrons(m_e*artificial_e_mass_multiplier, -1., n_0, vp_T_e, vn_T_e, const_colrate);
    electrons.description="electron distribution";

// // E
    E = new_nv ( MX ); /* Allocate u vector */
    q = new_nv ( MX ); /* Allocate u vector */

// // set data
    ions.E=electrons.E=&E;
    ions.electric_energy=electrons.electric_energy=&electric_energy;
    ions.other=&electrons;
    electrons.other=&ions;

    SetIC_electrons (electrons);
    SetIC_ions ( ions ); /* Initialize u vector */

    fprintf(stderr,"init i solver\n");
// // initialize ion solver
    /* Call CVodeCreate to create the solver memory and specify the
     * Backward Differentiation Formula and the use of a Newton iteration */
    ions.cvode_mem = CVodeCreate ( CV_BDF, CV_NEWTON );
    if ( check_flag ( ( void * ) ions.cvode_mem, "CVodeCreate", 0 ) ) {
        return ( 1 );
    }

    /* Call CVodeInit to initialize the integrator memory and specify the
     * user's right hand side function in u'=f(t,u), the inital time T0, and
     * the initial dependent variable vector u. */
    flag = CVodeInit ( ions.cvode_mem, f_collisionless<BGK>, T0, ions.f );
    if ( check_flag ( &flag, "CVodeInit", 1 ) ) {
        return ( 1 );
    }


    /* Call CVodeSStolerances to specify the scalar relative tolerance
     * and scalar absolute tolerance */
    flag = CVodeSStolerances ( ions.cvode_mem, RTOL, ATOL);
    if ( check_flag ( &flag, "CVodeSStolerances", 1 ) ) {
        return ( 1 );
    }

    /* Set the pointer to user-defined data_i */
    flag = CVodeSetUserData ( ions.cvode_mem, &ions );
    if ( check_flag ( &flag, "CVodeSetUserData", 1 ) ) {
        return ( 1 );
    }

    LS_i = SUNSPGMR(ions.f, PREC_NONE, 20);
    flag = CVSpilsSetLinearSolver(ions.cvode_mem,LS_i);
    if ( check_flag ( &flag, "GMRES", 1 ) ) {
        return ( 1 );
    }
//     ions.cvodememrec()->cv_hmin=1e-15;

//     CVodeSetMaxOrd( ions.cvode_mem, 2);
//     CVodeSetStabLimDet(ions.cvode_mem, 0);


    fprintf(stderr,"init e solver\n");
// // initialize ion solver
    /* Call CVodeCreate to create the solver memory and specify the
     * Backward Differentiation Formula and the use of a Newton iteration */
    electrons.cvode_mem = CVodeCreate ( CV_BDF, CV_NEWTON );
    if ( check_flag ( ( void * ) electrons.cvode_mem, "CVodeCreate", 0 ) ) {
        return ( 1 );
    }

    /* Call CVodeInit to initialize the integrator memory and specify the
     * user's right hand side function in u'=f(t,u), the inital time T0, and
     * the initial dependent variable vector u. */
    flag = CVodeInit ( electrons.cvode_mem, f_collisionless<BGK>, T0, electrons.f );
    if ( check_flag ( &flag, "CVodeInit", 1 ) ) {
        return ( 1 );
    }

    /* Call CVodeSStolerances to specify the scalar relative tolerance
     * and scalar absolute tolerance */
    flag = CVodeSStolerances ( electrons.cvode_mem, RTOL, ATOL);
    if ( check_flag ( &flag, "CVodeSStolerances", 1 ) ) {
        return ( 1 );
    }

    /* Set the pointer to user-defined data_i */
    flag = CVodeSetUserData ( electrons.cvode_mem, &electrons );
    if ( check_flag ( &flag, "CVodeSetUserData", 1 ) ) {
        return ( 1 );
    }

    LS_e = SUNSPGMR(electrons.f, PREC_NONE, 20);
    flag = CVSpilsSetLinearSolver(electrons.cvode_mem,LS_e);
    if ( check_flag ( &flag, "GMRES", 1 ) ) {
        return ( 1 );
    }
//     electrons.cvodememrec()->cv_hmin=1e-15;


//     fprintf(stderr,"init e solver\n");


//MAIN LOOP
    /* In loop over output points: call CVode, print results, test for errors */

    std::cerr << "set stuff" << std::endl;
    electrons.reserve(std::ceil(T1/DTOUT));
    ions.reserve(std::ceil(T1/DTOUT));

    std::cerr << electrons << std::endl;
    std::cerr << ions << std::endl;

    ions.update_n();
    realtype tmpnn=ions.n_total.back();

    electrons.update_n();
    q_from_ns(q, electrons, ions);
    E_from_q(E,q);
    export_to_netcdf(output_path+filename+"_0.nc", ions, electrons);

    realtype internal_step_i=0., internal_step_e=0.;

    realtype tout = DTOUT/2;

    for ( size_t iout=0; iout < NOUT && tout < T1; iout++, tout += DTOUT ) {

        mt_total.start();
        if (!one_step(tout, q, E, electric_energy, electrons, ions)){
            export_to_netcdf(output_path+filename+"_failure.nc", ions, electrons);
            return 1;
        }
        mt_total.stop();


        flag = CVodeGetNumSteps ( ions.cvode_mem, &ions.steps );
        check_flag ( &flag, "CVodeGetNumSteps", 1 );

//         CVodeGetCurrentOrder(ions.cvode_mem, &cv_order_i);
        CVodeGetCurrentStep(ions.cvode_mem, &internal_step_i );

        flag = CVodeGetNumSteps ( electrons.cvode_mem, &electrons.steps );
        CVodeGetCurrentStep(ions.cvode_mem, &internal_step_e );
        check_flag ( &flag, "CVodeGetNumSteps", 1 );

        std::cerr.precision(4);
        std::cerr
//         << std::string('\b',5) << std::string(" ",6)
        << "\ravg time=" << mt_total.get_mean() << "=" << mt_total.get_duration() << "/" << mt_total.count
        << ", last time=" << mt_total.get_last()
        << ", time left=" << mt_total.get_mean()*(total_steps-iout)
        << ", t=" << tout

        << ", io=" << io_from_rhs_time(mt_f.get_mean())
//         << ", t_i=" << ions.time
//         << ", cv_order_i=" << cv_order_i
        << ", cv_step_i=" << std::scientific << internal_step_i
        << ", cv_i%=" << std::scientific << mt_cvode_i.get_duration()/mt_total.get_duration()
//         << "(" << mt_cvode_i.count << ")"
        << ", nst_i=" << ions.steps
        << ", cv_step_e=" << std::scientific << internal_step_e
        << ", cv_e%=" << std::scientific << mt_cvode_e.get_duration()/mt_total.get_duration()
//         << "(" << mt_cvode_e.count << ")"
        << ", nst_e=" << electrons.steps // << std::endl;
        << ", rel_iloss=" << std::scientific << (ions.n_total.back()-tmpnn)/tmpnn
//         << ", nu_i=" << ions.colrate_total
//         << ", nu_e=" << electrons.colrate_total
        << "#";

        tmpnn=ions.n_total.back();

        size_t stops = 200;
        if (iout % (total_steps/stops) == 0 && iout > .005*total_steps && iout < .995*total_steps){
            size_t intermediate = std::floor((realtype)iout/total_steps*stops);

            export_to_netcdf(output_path+filename+"_"+std::to_string(intermediate)+".nc", ions, electrons);
        }

    }
    tout += DTOUT/2.;
    mt_total.start();
    if (!one_step(tout, q, E, electric_energy, electrons, ions)) return 1;
    mt_total.stop();

    std::cerr << std::endl;
    std::cout << mt_integrator << std::endl;
    std::cout << mt_efield << std::endl;
    std::cout << mt_f << std::endl;
    std::cout << mt_cvode_i << std::endl;
    std::cout << mt_cvode_e << std::endl;
    std::cout << mt_total << std::endl;

    export_to_netcdf(output_path+filename+"_final.nc", ions, electrons);

    destroy_nv ( q );  /* Free the u vector */
    destroy_nv ( E );  /* Free the u vector */
    CVodeFree ( &ions.cvode_mem ); /* Free the integrator memory */
    CVodeFree ( &electrons.cvode_mem ); /* Free the integrator memory */
    SUNLinSolFree ( LS_i );     /* Free the linear solver memory */
    SUNLinSolFree ( LS_e );     /* Free the linear solver memory */

    return ( 0 );
}


