#pragma once

#include <cmath>
#include <exception>
#include <limits>

#include "openmp_nv_overlay.h"
#include "cvode/cvode_impl.h"

struct UserData{

    UserData(realtype _mass,
             realtype _charge,
             realtype _n_nominal,
             realtype _vp_T_nominal,
             realtype _vn_T_nominal,
             realtype colrate = 0.) : m(_mass), q(_charge), n_nominal(_n_nominal/n_normalization), vn_T_nominal(_vn_T_nominal/v_normalization), vp_T_nominal(_vp_T_nominal/v_normalization),
                            vnmin(0.), vnmax(+vn_T_nominal*domain_thermal_factor), dvn((vnmax-vnmin)/(mvn+(boundary_vn!=btype::PERIODIC))), rezi_dvn(1./dvn),
                            vpmin(-vp_T_nominal*domain_thermal_factor), vpmax(+vp_T_nominal*domain_thermal_factor), dvp((vpmax-vpmin)/(mvp+(boundary_vp!=btype::PERIODIC))), rezi_dvp(1./dvp),
                            vnstart(vnmin+(boundary_vn!=btype::PERIODIC)*dvn), vpstart(vpmin+(boundary_vp!=btype::PERIODIC)*dvp),
                            advection_coeffi(compute_advection_coeffi()),
                            const_colrate(colrate/X_LENGTH*std::sqrt(vn_T_nominal*vn_T_nominal+ vp_T_nominal*vp_T_nominal)),
                            null_vp(vpmin>0?0:(-vpmin/dvp))
                            {
                                f = new_nv ( mx*mvn*mvp );
                                n.resize(mx);
                                v_thermal_p.resize(mx);
                                v_thermal_n.resize(mx);
                                v_thermal.resize(mx);
                                col_rate.resize(mx);
                                u_n.resize(mx);
                                u_p.resize(mx);
                            };

    ~UserData() {
        destroy_nv(f);
    }

    constexpr static size_t mx = MX, mvn = MVn, mvp = MVp;
    const realtype m, q, n_nominal, vn_T_nominal, vp_T_nominal;
    constexpr static realtype xmin = XMIN, xmax = XMAX, dx=DX, rezi_dx=1./DX, xstart=XSTART;
    const realtype vnmin, vnmax, dvn, rezi_dvn, vnstart;
    const realtype vpmin, vpmax, dvp, rezi_dvp, vpstart;
    const realtype normalization_speed = v_normalization;

    const realtype advection_coeffi;
    const realtype const_colrate;
    const size_t null_vp;

    N_Vector f = NULL;

    std::string description;

    N_Vector* E = NULL;
    std::vector<realtype>* electric_energy = NULL;
    UserData* other = NULL;

    // x dependent data
    std::vector<realtype> n;
    std::vector<realtype> v_thermal_p;
    std::vector<realtype> v_thermal_n;
    std::vector<realtype> v_thermal;
    std::vector<realtype> col_rate;
    std::vector<realtype> u_n;
    std::vector<realtype> u_p;

    // evolution of observables
    std::vector<realtype> n_total = {};
    std::vector<realtype> kinetic_energy = {};
    std::vector<realtype> momentum_flux_r = {};
    std::vector<realtype> momentum_flux_l = {};
    std::vector<realtype> heat_flux = {};
    std::vector<realtype> energy_flux = {};
    std::vector<realtype> temp_p = {};
    std::vector<realtype> temp_n = {};

    realtype x(size_t i) const { return xstart +i*dx; }
    realtype vp(size_t i) const { return vpstart +i*dvp; }
    realtype vn(size_t i) const { return vnstart +i*dvn; }

    void update_n();
    void compute_heat_flux(const size_t pos = MX-1);

    // algorithm specific mn_data
    void* cvode_mem = NULL;
    // NOTE: CVodeMemRec is described with its memeber in cvode_impl.h
    CVodeMemRec* cvodememrec() {
        return (CVodeMemRec*) cvode_mem;
    }

    long steps = 0;
    realtype time = 0.;

    void reserve(size_t size){
        n_total.reserve(size);
        kinetic_energy.reserve(size);
        electric_energy->reserve(size);
        temp_p.reserve(size);
        temp_n.reserve(size);
        heat_flux.reserve(size);
        energy_flux.reserve(size);
        momentum_flux_r.reserve(size);
        momentum_flux_l.reserve(size);
    }

    realtype colrate_total = 0.;
    std::string print() const;

private:
    realtype compute_advection_coeffi(){

        return n_nominal*q*unit_m/m;
    }
};

std::string UserData::print() const {
    std::ostringstream out;

    out << "UserData: " << description << ":" << std::endl;
    out << "\t" << "mass=\t" << m << std::endl;
    out << "\t" << "charge=\t" << q << std::endl;
    out << "\t" << "n_nominal=\t" << n_nominal*n_normalization << "(" << n_nominal << ")" << std::endl;
    out << "\t" << "v_normalization=\t" << normalization_speed << std::endl;
    out << "\t" << "vp_T=\t" << vp_T_nominal << std::endl;
    out << "\t" << "vn_T=\t" << vn_T_nominal << std::endl;
    out << "\t" << "L=lambda_D_0=\t" << lambda_D_0 << std::endl;
    out << "\t" << "tau=\t" << tau << std::endl;
    out << "\t" << "pf=\t" << advection_coeffi << std::endl;
    out << "\t" << "const_colrate=\t" << const_colrate << std::endl;
//     out << "\t" << "null_vp=" << null_vp << ", vp(null_vp)=" << vp(null_vp)<< std::endl;

    return out.str();
}

std::ostream& operator<<(std::ostream& out, const UserData& input){
    out << input.print();
    return out;
}


void UserData::update_n()
{
    realtype* udata = N_VGetArrayPointer(f);

    const realtype d_factor = dvn*dvp;
    realtype kin_energy = 0.;
    realtype n_sum = 0.;

    colrate_total = 0;

#pragma omp parallel for reduction(+:n_sum,kin_energy,colrate_total)
    for (size_t iter_x=0; iter_x < mx; iter_x++) {
        realtype tmp_n=0.;
        realtype tmp_vn=0.;
        realtype tmp_vp=0.;
        realtype un = 0.;
        realtype up = 0.;


// compute density, drifts and kinetic energy
        realtype vn_l = vn(0);
        for (size_t iter_vn=0; iter_vn < mvn; iter_vn++, vn_l+=dvn) {
            realtype vp_l = vp(0);
            for (size_t iter_vp=0; iter_vp < mvp; iter_vp++, vp_l+=dvp) {
                realtype uijk = IJKth(udata,iter_x,iter_vn,iter_vp );

                tmp_n += uijk;
//                 un += uijk * vn_l;
                up += uijk * vp_l;
                tmp_vn += uijk * std::pow(vn_l,2);
                tmp_vp += uijk * std::pow(vp_l,2);
            }
        }
        n[iter_x]=tmp_n*d_factor;
        n_sum+=n[iter_x];
        kin_energy+=(tmp_vp+tmp_vn);

//         un *= d_factor;
        u_n[iter_x] = un = 0.;
        up *= d_factor;
        u_p[iter_x] = up;

        if (iter_x==MX-1) {
            realtype upl=0.;
            for (size_t iter_vn=0; iter_vn < mvn; iter_vn++, vn_l+=dvn) {
                realtype vp_l = vp(null_vp);
                for (size_t iter_vp=null_vp; iter_vp < mvp; iter_vp++, vp_l+=dvp) {
                    realtype uijk = IJKth(udata,iter_x,iter_vn,iter_vp );
                    upl += uijk * vp_l;
                }
            }
            momentum_flux_r.push_back(upl*d_factor);
        } else if (iter_x==0) {
            realtype upl=0.;
            for (size_t iter_vn=0; iter_vn < mvn; iter_vn++, vn_l+=dvn) {
                realtype vp_l = vp(0);
                for (size_t iter_vp=0; iter_vp < null_vp; iter_vp++, vp_l+=dvp) {
                    realtype uijk = IJKth(udata,iter_x,iter_vn,iter_vp );
                    upl += uijk * vp_l;
                }
            }
            momentum_flux_l.push_back(-upl*d_factor);
        }

        realtype tmp_temp_n = 0.;
        realtype tmp_temp_p = 0.;
        vn_l = vn(0);
        for (size_t iter_vn=0; iter_vn < mvn; iter_vn++, vn_l+=dvn) {
            realtype vn_l_2 = std::pow(vn_l-un,2);
            realtype vp_l = vp(0);
            for (size_t iter_vp=0; iter_vp < mvp; iter_vp++, vp_l+=dvp) {
                realtype uijk = IJKth(udata,iter_x,iter_vn,iter_vp );

                tmp_temp_p += uijk*std::pow((vp_l-up),2);
                tmp_temp_n += uijk*vn_l_2;
            }
        }

        if(n[iter_x]<std::numeric_limits<realtype>::epsilon()*1.e5){
//             std::clog << "n (" << n[iter_x] << ") too small at " << iter_x << std::endl;
            n[iter_x]=0.;
            v_thermal_p[iter_x] = 0.;
            v_thermal_n[iter_x] = 0.;
            v_thermal[iter_x] = 0.;
        } else {
            v_thermal_p[iter_x] = std::sqrt(tmp_temp_p * d_factor/n[iter_x]);
            v_thermal_n[iter_x] = std::sqrt(tmp_temp_n * d_factor/n[iter_x]);

            realtype pf = 2.;
            if (std::isnan(v_thermal_p[iter_x]) || v_thermal_p[iter_x]<pf*dvp) {
//                 std::cerr << "Limiting vTp (" << v_thermal_p[iter_x] << ") with n=" << n[iter_x] << " at " << iter_x << std::endl;
                v_thermal_p[iter_x]=dvp*pf;
            }
            if (std::isnan(v_thermal_n[iter_x]) || v_thermal_n[iter_x]<pf*dvn) {
//                 std::cerr << "Limiting vTn (" << v_thermal_n[iter_x] << ") with n=" << n[iter_x] << " at " << iter_x <<  std::endl;
                v_thermal_n[iter_x]=dvn*pf;
            }

            v_thermal[iter_x] = std::sqrt((std::pow(v_thermal_p[iter_x],2)+std::pow(v_thermal_n[iter_x],2))/3.);

        }


        // heat + energy flux @ last point
        if (iter_x == MX-1) {
            heat_flux.push_back(0.);
            energy_flux.push_back(0.);
            for (size_t iter_vn=0; iter_vn < mvn; iter_vn++) {
                realtype vn_l_2 = std::pow(vn( iter_vn )-un,2);
                realtype vn_l_2_e = std::pow(vn( iter_vn ),2);
                for (size_t iter_vp=0; iter_vp < mvp; iter_vp++) {
                    realtype vp_l = vp( iter_vp );
                    realtype uijk = IJKth(udata,iter_x,iter_vn,iter_vp );

                    heat_flux.back() += uijk*(std::pow(vp_l-up,2)+vn_l_2) * vp_l;
                    energy_flux.back() += uijk*(std::pow(vp_l,2)+vn_l_2) * vp_l;
                }
            }
            heat_flux.back() *= d_factor;
            energy_flux.back() *= d_factor;
        }

        // temperature in the middle
        if (iter_x == std::floor(mx/2)) {
            temp_p.push_back(std::pow(v_thermal_p[iter_x],2)*.5*m );
            temp_n.push_back(std::pow(v_thermal_n[iter_x],2)*.5*m );
        }
    }
    colrate_total/=mx;
    n_total.push_back(n_sum*dx);
    kinetic_energy.push_back(kin_energy*.5*m*d_factor*dx);
}
